import os
import sys
import json
from getpass import getpass

from fuzzywuzzy.fuzz import ratio
from nacl.exceptions import CryptoError

from .util import error
from .crypto import encrypt, decrypt

# Should match at least this much ratio with fuzzy string matching.
RATIO = 85


class Wheel:
    def __init__(self):
        self.wheel = None
        self.load_or_create_wheel()

    @property
    def path(self):
        return os.path.expanduser('~/.passwheel')

    def load_or_create_wheel(self):
        if os.path.exists(self.path):
            with open(self.path, 'rb') as f:
                self.wheel = f.read()
        else:
            self.wheel = self.create_wheel()

    def create_wheel(self):
        print(
            '{} doesnt exist, initializing.\n'
            'Please create master password.'.format(self.path)
        )
        pw = self.get_pass(prompt='master password: ', verify=True)
        self.wheel = self.encrypt_wheel({}, pw)

    def get_unlock_pw(self):
        return self.get_pass(prompt='unlock: ')

    def get_pass(self, prompt=None, verify=False):
        if not prompt:
            prompt = 'password: '
        while True:
            pw = getpass(prompt).encode('utf8')
            if not verify:
                return pw
            pw2 = getpass('verify {}'.format(prompt)).encode('utf8')
            if pw == pw2:
                return pw
            print('password mismatch')

    def decrypt_wheel(self, pw):
        try:
            plaintext = decrypt(pw, self.wheel)
        except CryptoError:
            error('unlock failed!')
            sys.exit(1)
        return json.loads(plaintext.decode('utf8'))

    def encrypt_wheel(self, data, pw):
        plaintext = json.dumps(data).encode('utf8')
        ciphertext = encrypt(pw, plaintext)
        with open(self.path, 'wb') as f:
            f.write(ciphertext)
        return ciphertext

    def add_login(self, service, username, password):
        pw = self.get_unlock_pw()
        data = self.decrypt_wheel(pw)
        data[service] = data.get(service) or {}
        if isinstance(password, bytes):
            password = password.decode('utf8')
        data[service][username] = password
        self.wheel = self.encrypt_wheel(data, pw)

    def change_password(self):
        pw = self.get_unlock_pw()
        data = self.decrypt_wheel(pw)
        new_pw = self.get_pass(prompt='new master password: ', verify=True)
        self.wheel = self.encrypt_wheel(data, new_pw)

    def get_login(self, service):
        pw = self.get_unlock_pw()
        data = self.decrypt_wheel(pw)
        return data.get(service) or {}

    def find_login(self, query):
        pw = self.get_unlock_pw()
        data = self.decrypt_wheel(pw)
        ratios = sorted(
            (
                (ratio(query, key), key)
                for key in data
            ),
            reverse=True,
        )
        threshold = min(RATIO, ratios[0][0])
        top = [
            key
            for r, key in ratios
            if r >= threshold
        ]
        return [
            (key, data[key])
            for key in top
        ]

    def rm_login(self, service, login):
        pw = self.get_unlock_pw()
        data = self.decrypt_wheel(pw)
        if login is None:
            del data[service]
        else:
            del data[service][login]
            if not data[service]:
                del data[service]
        self.wheel = self.encrypt_wheel(data, pw)



'''
passwheel

A password and secret personal storage tool.
'''

__title__ = 'passwheel'
__version__ = '0.1.2'
__all__ = ('Wheel',)
__author__ = 'Johan Nestaas <johannestaas@gmail.com>'
__license__ = 'GPLv3'
__copyright__ = 'Copyright 2019 Johan Nestaas'

import sys

from .wheel import Wheel
from .passgen import gen_password
from .clipboard import copy
from .util import info, warning, error


def main():
    import argparse
    parser = argparse.ArgumentParser()
    subs = parser.add_subparsers(dest='cmd')

    p = subs.add_parser('add', help='add a login')
    p.add_argument('service', help='service/website')
    p.add_argument('username', help='login')
    p.add_argument(
        '--custom', '-c', action='store_true',
        help='input custom password',
    )
    p.add_argument(
        '--words', '-w', type=int, default=2,
        help='number of words in generated password',
    )
    p.add_argument(
        '--digits', '-d', type=int, default=3,
        help='number of digits in generated password',
    )
    p.add_argument(
        '--symbol', '-s', action='store_true',
        help='append a random symbol to the password',
    )

    p = subs.add_parser('rm', help='remove a service or login')
    p.add_argument('service', help='service/website')
    p.add_argument('username', nargs='?', default=None, help='login')

    p = subs.add_parser('get', help='fetch creds for service/website')
    p.add_argument('service', help='service/website')
    p.add_argument('username', nargs='?', default=None, help='login name')
    p.add_argument(
        '--copy', '-c', action='store_true',
        help='copy to clipboard',
    )

    p = subs.add_parser('find', help='find creds with fuzzy matching')
    p.add_argument('query', help='query for service/website')

    p = subs.add_parser('dump', help='dump all decrypted credentials')
    p.add_argument('service', nargs='?', default=None, help='service/website')
    p.add_argument(
        '--no-passwords', '-n', action='store_true',
        help='dont print passwords',
    )

    p = subs.add_parser('changepw', help='change master password')

    args = parser.parse_args()

    wheel = Wheel()
    if args.cmd == 'dump':
        pw = wheel.get_pass(prompt='unlock: ')
        data = wheel.decrypt_wheel(pw)
        if args.service:
            data = {
                k: v
                for k, v in data.items()
                if k.lower() == args.service.lower()
            }
        if not data:
            warning('no passwords found.')
        else:
            for service, logins in data.items():
                print(service)
                for user, pw in logins.items():
                    if args.no_passwords:
                        print('  {}'.format(user))
                    else:
                        print('  {}: {}'.format(user, pw))
    elif args.cmd == 'add':
        if args.custom:
            add_pw = wheel.get_pass(prompt='new password: ', verify=True)
        else:
            add_pw, entropy = gen_password(
                num_words=args.words,
                num_digits=args.digits,
                add_symbol=args.symbol,
            )
            info('generated password with {} bits of entropy'.format(
                int(entropy),
            ))
            if copy(add_pw):
                info('password copied to clipboard')
            else:
                warning("couldn't copy to clipboard, please run `get` command")
        wheel.add_login(args.service, args.username, add_pw)
    elif args.cmd == 'rm':
        wheel.rm_login(args.service, args.username)
    elif args.cmd == 'find':
        for service, logins in wheel.find_login(args.query):
            print('service: {}'.format(service))
            for key, val in logins.items():
                print('  {}|{}'.format(key, val))
    elif args.cmd == 'get':
        logins = sorted(wheel.get_login(args.service).items())
        if args.username:
            logins = [
                (user, pw)
                for user, pw in logins
                if user == args.username
            ]
        if not logins:
            warning('{!r}::{!r} not found'.format(
                args.service,
                args.username or '*',
            ))
            sys.exit(1)
        if args.copy:
            if len(logins) > 1:
                users = sorted([x[0] for x in logins])
                warning(
                    '{} logins found, please specify user from: {}'
                    .format(len(logins), ', '.join(users))
                )
                sys.exit(2)
            user, pw = logins[0]
            if copy(pw):
                info('copied password to clipboard for user:')
                print(user)
            else:
                error('couldnt copy password to clipboard. Please remove -c.')
                sys.exit(3)
        else:
            if args.username:
                # Just print the first and only password.
                print(logins[0][1])
            else:
                # Dump the users and passwords for this service.
                for key, val in logins:
                    print('{}|{}'.format(key, val))
    elif args.cmd == 'changepw':
        wheel.change_password()
    else:
        parser.print_usage()


if __name__ == '__main__':
    main()



import os
import sys
from colorama import Fore
from subprocess import check_output


def has_binary(binary):
    return bool(check_output(['which', binary]).strip())


def get_os():
    '''
    passwheel is officially targeted for linux systems.
    This returns "linux" or otherwise.
    '''
    return os.uname().sysname.lower()


def info(s):
    print(s, file=sys.stderr)


def warning(s):
    print('{}{}{}'.format(Fore.YELLOW, s, Fore.RESET), file=sys.stderr)


def error(s):
    print('{}{}{}'.format(Fore.RED, s, Fore.RESET), file=sys.stderr)



from subprocess import Popen, PIPE

from .util import has_binary, warning, get_os


def copy(s):
    if isinstance(s, str):
        s = s.encode('utf8')
    if get_os() == 'linux':
        if not has_binary('xclip'):
            warning('xclip not installed, cant modify clipboard')
            return False
        Popen(['xclip', '-selection', 'clipboard'], stdin=PIPE).communicate(s)
        return True
    elif get_os() == 'darwin':
        if not has_binary('pbcopy'):
            warning('pbcopy not installed, cant modify clipboard')
            return False
        Popen(['pbcopy'], stdin=PIPE).communicate(s)
        return True
    else:
        warning('passwheel only supports linux for modifying the clipboard')
    return False



import math
import random

SYMBOLS = '!@$%^&*='


with open('/usr/share/dict/words') as f:
    WORDS = {
        x.strip().lower().replace("'s", '')
        for x in f
    }
    WORDS = sorted({
        x for x in WORDS
        if 3 < len(x) < 12
    })


def get_word():
    return random.choice(WORDS)


def gen_password(num_words=2, num_digits=3, add_symbol=False):
    words = [get_word() for i in range(num_words)]
    if num_digits > 0:
        digits_fmt = '{{:0{}}}'.format(num_digits)
        digits = digits_fmt.format(random.randint(0, 10**num_digits - 1))
    else:
        digits = ''
    i = random.randint(0, num_words - 1)
    words[i] = words[i].upper()
    if add_symbol:
        symbol = random.choice(SYMBOLS)
    else:
        symbol = ''
    entropy = len(WORDS)**num_words
    entropy *= 10**num_digits
    if add_symbol:
        entropy *= len(SYMBOLS)
    entropy = math.log(entropy, 2)
    pw = ''.join(words + [digits] + [symbol])
    return pw, entropy



import nacl.pwhash
import nacl.secret
import nacl.utils


def kdf(password):
    salt = b'PASSPASSWHEELEEE'
    return nacl.pwhash.argon2i.kdf(
        nacl.secret.SecretBox.KEY_SIZE,
        password,
        salt,
        opslimit=nacl.pwhash.argon2i.OPSLIMIT_SENSITIVE,
        memlimit=nacl.pwhash.argon2i.MEMLIMIT_SENSITIVE,
    )


def encrypt(password, message):
    key = kdf(password)
    box = nacl.secret.SecretBox(key)
    return box.encrypt(message)


def decrypt(password, message):
    key = kdf(password)
    box = nacl.secret.SecretBox(key)
    return box.decrypt(message)
